#! /usr/bin/env python
# 
# gdmapriori: GNOME interface to plot data files
#
# TODO
#
#	How to simulate a click on the Browse button to connect
#	this to the Open callbacks
#
# 	Better structure of generate_plot
#
#	Handle plots of string frequencies
#	

from gnome.ui import *

import gtk
import libglade
import os
import sys
import string
import re
import tempfile
import signal

#import Gnuplot

#import dmtools

programName = os.path.basename (sys.argv[0])

# Record the name of the currently open dataset so that we only need
# load it once.

loadedDataFile = ""
loadedDataSet  = []

# Record the list of process id's for all the external plots.

plotidList = []

# Record the default value of tempdir so that I can change back to it
# when required.

defaultTempDir = tempfile.tempdir

def readDataFile(dataFileName):
    """Read a transaction dataset from the dataFileName.
The dataFileName is a string naming a file containing the transaction data.
A dataset object (list of lists corresponding to a records of transactions)
is returned."""

    global loadedDataSet
    global loadedDataFile

    # Check to see if this is a new dataset file name and if not
    # simply return the dataset already loaded.

    if loadedDataFile == dataFileName: return (loadedDataSet)
	
    # Try openning the data file and throw an exception if this
    # fails.
	
    try:
        dataFile = open (dataFileName, "r")
    except IOError, e:
        message = programName + ": readDataFile" + str (e)
	raise IOError, message
		
    print "Reading data from the file \"%s\"" % (dataFileName)

    inputLine = dataFile.readline ()
    ignoredLines = 0
    recordCount = 0
    loadedDataSet = []

    while inputLine and (inputLine[0] == '#' or \
			 re.match (r'^$', inputLine)):
	ignoredLines = ignoredLines + 1
	#print "Skip '%s'" % (inputLine)
	inputLine = dataFile.readline ()
		
    # Try to guess the datatypes.  This is used to convert what is
    # read to float or integer, or else to leave it as a string.
        
    inputLine = string.strip (inputLine)
    itemList = string.split (inputLine)
    datatypes = []
    for i in range (len(itemList)):
	if re.match (r'^\D', itemList[i]):
	    #print "Field %d is non-digit: %s" % (i, itemList[i])
	    datatypes.append ("string")
	elif re.match (r'^\d+$', itemList[i]):
	    #print "Field %d is integer: %s" % (i, itemList[i])
	    datatypes.append ("integer")
	elif re.match (r'^\d+\.\d+$', itemList[i]):
	    #print "Field %d is float: %s" % (i, itemList[i])
	    datatypes.append ("float")
	else:
	    print "Field %d is NOT RECOGNISED: %s" % (i, itemList[i])
	    print "THIS SHOULD RAISE AN EXCEPTION SOMEDAY"

    # Read each input line, skipping any that begin with a `#' and
    # all blank lines.
	
    while inputLine:
	if inputLine[0] == '#' or re.match (r'^$', inputLine):
	    ignoredLines = ignoredLines + 1
	else:
	    recordCount = recordCount + 1
	    inputLine = string.strip (inputLine)
	    inputList = string.split (inputLine)
	    itemList = []
	    for i in range (len (inputList)):
		strippedItem = string.strip (inputList[i])
		if datatypes[i] == "string":
		    convertedItem = strippedItem
		elif datatypes[i] == "integer":
		    # First convert to float in case it might
		    # be a .0 or something else and the first
		    # line was an integer!
		    convertedItem = int(string.atof (strippedItem))
		elif datatypes[i] == "float":
		    convertedItem = string.atof (strippedItem)
		itemList = itemList + [convertedItem]
	    loadedDataSet = loadedDataSet + [itemList]
	inputLine = dataFile.readline ()

    dataFile.close ();

    # Record the name of the file just loaded so we can avoid
    # multiple loads.
	
    loadedDataFile = dataFileName

    print "There were %d records" % (recordCount)

    return loadedDataSet

def clear_all_plots (obj):

    # Currently, until I get the forking working, simply grab the list
    # of process ID's to kill and hope you don't mind killing
    # non-related gnuplot runs! This will continue even if you try to
    # kill someone elses gnuplot and fail.

    idList = os.popen ("/bin/ps -a | /bin/grep gnuplot | /usr/bin/awk '{printf $1\" \"}'", 'r').readline ()
    #print idList
    plotidList = string.split (idList)
    #print plotidList

    for i in plotidList:
        try:
	    os.kill ( string.atoi (i), signal.SIGTERM)
	except:
	    pass
	    #print "Ignore process %d\n" % (string.atoi (i))
	      

def on_about_activate (obj):
    "Callback: Display the about dialog."
    about = libglade.GladeXML ("gdmplot.glade", "about_dialog").\
            get_widget ("about_dialog")
    about.show ()

def on_subset_checkbutton_toggled (obj):
    subsetCheckButton = main_window.get_widget ('subset_checkbutton')
    subsetEntry =  main_window.get_widget ('subset_entry')
    subsetEntry.set_sensitive (subsetCheckButton.get_active ())

def popup_file_selection (obj):
    dataset_entry = main_window.get_widget ('dataset_entry')
    # How to get the browse button.  If I can, perhaps I can "Click" it here

########################################################################
# AXIS LABELS DIALOG
########################################################################

def popup_axis_labels_dialog (obj):
    global axis_labels_dialog
    axis_labels_dialog = libglade.GladeXML ("gdmplot.glade",
					    "axis_labels_dialog")
    dic = {"on_axis_labels_ok_button_clicked"     :
	   on_axis_labels_ok_button_clicked,
	   "on_axis_labels_apply_button_clicked"  :
	   on_axis_labels_apply_button_clicked,
	   "on_axis_labels_cancel_button_clicked"  :
	   on_axis_labels_cancel_button_clicked}
    axis_labels_dialog.signal_autoconnect (dic)
    axis_labels_dialog.get_widget ('x_axis_entry').set_text (xlabel)
    axis_labels_dialog.get_widget ('y_axis_entry').set_text (ylabel)

def on_axis_labels_apply_button_clicked (obj):
    global xlabel
    global ylabel
    xlabel = axis_labels_dialog.get_widget ('x_axis_entry').get_text ()
    ylabel = axis_labels_dialog.get_widget ('y_axis_entry').get_text ()

def on_axis_labels_ok_button_clicked (obj):
    global xlabel
    global ylabel
    xlabel = axis_labels_dialog.get_widget ('x_axis_entry').get_text ()
    ylabel = axis_labels_dialog.get_widget ('y_axis_entry').get_text ()
    axis_labels_dialog.get_widget ('axis_labels_dialog').close ()

def on_axis_labels_cancel_button_clicked (obj):
    axis_labels_dialog.get_widget ('axis_labels_dialog').close ()

########################################################################

def quit_app (obj):
    clear_all_plots (obj)
    gtk.mainquit ()

def generate_plot (obj):

    # Obtain from the GUI the name of the dataset and read in the
    # actual data.
    
    dataFileName = main_window.get_widget ('dataset_entry').gtk_entry ()\
		   .get_text ()
    try:
        dataset = readDataFile (dataFileName)
    except IOError, e:
        libglade.GladeXML ("gdmplot.glade", "filenotfound_messagebox")
        return
    
    #print dataset
    #print dataFileName

    # Obtain the column and number of bins from the GUI.

    columnNumber = main_window.get_widget ('column_spinbutton')\
		   .get_value_as_int ()
    binCount = main_window.get_widget ('bin_spinbutton').get_value_as_int ()
    #print "Plot column %d of %s in %d bins" % \
    #      (columnNumber, dataFileName, binCount)
    
    plotdata = []
    for i in range (0, len (dataset)):
        plotdata.append (dataset[i][columnNumber-1])
    #print plotdata
    minValue = min (plotdata)
    maxValue = max (plotdata)

    # If the column contains integers 1 - 7.  If I select 7 bins the
    # bins do not align with the integers!  This is correct because
    # there are only 6 bins between 1 and 7. If I want the bins
    # aligned with the integers I would choose 6 bins! But then
    # again?????  So I added 1.0 in the equation below.  I have not
    # thought through the consequences on continuous data!!!!!!  Looks
    # okay though.

    binSize = ((1.0 + float(maxValue) - float(minValue))/float(binCount))
    #print "Min = %s, Max = %s, Bins = %d, BinSize = %f, Mult = %f" %\
    #      (minValue, maxValue, binCount, binSize, (binSize*binCount)+minValue)

    # Generate the histogram after checking if subsetting is turned
    # on. If subsetting is on then only include in the histogram those
    # records that meet the condition. Note that an empty condition is
    # equivalent to turning subsetting off.

    subsetToggleButton = main_window.get_widget ('subset_checkbutton')
    subsetCondition = main_window.get_widget ('subset_entry'). get_text ()
    doSubset = subsetToggleButton.active and \
	       not re.match(r'^$', subsetCondition)
    if doSubset:
        patternMatch = re.search (r'^C(\d+)=(.*)$', subsetCondition, \
                                  re.IGNORECASE)
        #print subsetCondition

        if patternMatch:
            #print patternMatch.group (1)
            #print patternMatch.group (2)
            subsetColumn = string.atoi (patternMatch.group (1))
            subsetValue  = patternMatch.group (2)
        else:
            print "Error in format of condition string: SHOULD THROW EXCEPTION"
    
    binHistogram = {}
    recordCount = 0
    for i in range (len (plotdata)):
        if not (doSubset and str (dataset[i][subsetColumn-1]) != subsetValue):

            recordCount = recordCount + 1
            binIndex = int ((plotdata[i] - minValue)/binSize)
            if binHistogram.has_key (binIndex):
                frequency = binHistogram[binIndex] + 1
            else:
                frequency = 1
            #print "(%s,%s)" % (plotdata[i], binIndex)
            binHistogram.update ({binIndex:frequency})
        
    #print binHistogram

    # Generate temporary files in current directory (to avoid security
    # questions) and later delete them.

    if main_window.get_widget ('local_temp_files_checkbutton').active:
	tempfile.tempdir = "."
    else:
	tempfile.tempdir = defaultTempDir
    plotDataFilename = tempfile.mktemp(".data")
    plotDataFile = open (plotDataFilename, "w")
    #print plotDataFilename
    for k in binHistogram.keys ():
        plotDataFile.write (str (minValue+(k*binSize)) + " " + \
                            str (binHistogram[k]) + "\n")
    plotDataFile.close ()

    legend = "Plot of " + os.path.basename (dataFileName) + \
	     " column " + str (columnNumber) + " into " + \
	     str (binCount) + " bins"
    if doSubset:
	    legend = legend + " using subset " + subsetCondition
    legend = legend + " with " + str (recordCount) + " records"
    print legend
    
    scriptFileName = tempfile.mktemp(".gnuplot")
    scriptFile = open (scriptFileName, "w")
    #script_init = "set key below\n"
    script_init = "set nokey\n"
    script_init = script_init + "set xlabel \"%s\"\n" % (xlabel)
    script_init = script_init + "set ylabel \"%s\"\n" % (ylabel)
    scriptFile.write (script_init)

    gnuplot_plot_command = "plot " + \
                           "[" + str (minValue-binSize) + \
                           ":" + str (maxValue+binSize) + \
                           "] "  + \
                           "[0:] '" + plotDataFilename + "' "  \
                           "title \"" + legend + "\""  \
                           " with impulses linewidth 20\n"


    if main_window.get_widget ('window_checkbutton').active:
	script_x11 = "set term x11\n" \
		    "set title \"" + legend + "\"\n"
	scriptFile.write (script_x11)
	scriptFile.write (gnuplot_plot_command)

    if main_window.get_widget ('latex_checkbutton').active:
	latexFileName = re.sub ('\.', '-', os.path.basename (dataFileName))
	latexFileName = latexFileName + \
			"-C" + str (columnNumber) + "B" + str (binCount)
	if doSubset:
	    latexFileName = latexFileName + "-" + subsetCondition
	latexFileName = latexFileName + 	".tex"

	scriptLaTeX = "set term latex\n"
      	if main_window.get_widget ('title_latex_checkbutton').active:
	    scriptLaTeX = scriptLaTeX + \
			  "set title \"" + legend + "\"\n" 
	else:
	    scriptLaTeX = scriptLaTeX + \
			  "set title\n" 
	scriptLaTeX = scriptLaTeX + \
		      "set output \'" + latexFileName + "\'\n"
	scriptFile.write (scriptLaTeX)
	scriptFile.write (gnuplot_plot_command)
	
    if main_window.get_widget ('postscript_checkbutton').active:
	postscriptFileName = re.sub ('\.', '-',os.path.basename (dataFileName))
	postscriptFileName = postscriptFileName + \
			"-C" + str (columnNumber) + "B" + str (binCount)
	if doSubset:
	    postscriptFileName = postscriptFileName + "-" + subsetCondition
	postscriptFileName = postscriptFileName + 	".eps"

	scriptPostScript = "set term postscript eps monochrome solid 20\n" 
	if main_window.get_widget ('title_eps_checkbutton').active:
	    scriptPostScript = scriptPostScript + \
			       "set title \"" + legend + "\"\n" 
	else:
	    scriptPostScript = scriptPostScript + \
			       "set title\n" 

	scriptPostScript = scriptPostScript + \
			   "set output \'" + postscriptFileName + "\'\n"
	scriptFile.write (scriptPostScript)
	scriptFile.write (gnuplot_plot_command)

    if main_window.get_widget ('pdf_checkbutton').active:
	epsFileName = re.sub ('\.', '-',os.path.basename (dataFileName))
	epsFileName = epsFileName + \
			"-C" + str (columnNumber) + "B" + str (binCount)
	if doSubset:
	    epsFileName = epsFileName + "-" + subsetCondition
	epsFileName = epsFileName + ".teps"

	scriptCommands = "set term postscript eps monochrome solid 20\n" 
	if main_window.get_widget ('title_pdf_checkbutton').active:
	    scriptCommands = scriptCommands + \
			       "set title \"" + legend + "\"\n" 
	else:
	    scriptCommands = scriptCommands + \
			       "set title\n" 

	scriptCommands = scriptCommands + \
			   "set output \'" + epsFileName + "\'\n"
	scriptFile.write (scriptCommands)
	scriptFile.write (gnuplot_plot_command)

    if main_window.get_widget ('printer_checkbutton').active:
	printerFileName = tempfile.mktemp(".ps")
	scriptPrinter = "set term postscript default color solid 20\n" \
			"set title \"" + legend + "\"\n" \
			"set output \'" + printerFileName + "\'\n"
	scriptFile.write (scriptPrinter)
	scriptFile.write (gnuplot_plot_command)

    scriptFile.close ()

    os.system ("gnuplot -persist " + scriptFileName)

    if main_window.get_widget ('printer_checkbutton').active:
	os.system ("lpr " + printerFileName)
	os.remove (printerFileName)

    if main_window.get_widget ('pdf_checkbutton').active:
        os.system ("epstopdf " + epsFileName)
	os.remove (epsFileName)

# I could not get the following to work? The execvp seems to run once
# and then gnuplot gets confused next time. I'm probably
# confused. Also I'm confused about forking et al.

#    plotid = os.fork()
#    if not plotid:
#	os.execvp('gnuplot', ['gnuplot', '-persist', scriptFileName])
#	#os.execvp('gedit', [])

#    global plotidList
#    plotidList = plotidList + [plotid]
#    print plotidList
    
    os.remove (plotDataFilename)
    os.remove (scriptFileName)
    
#    plot=os.system("graph -T pnm test.data | xv -")
#    plot=os.system("graph -T X test.data")
#    print plot

#    gplot = Gnuplot.Gnuplot (debug=1)
#    gplot.title ("Testing")
#    gplot.plot ([[1,23],[2,24],[3,25],[4,26],[5,27]])
#    raw_input('Please press return to continue...\n')



#    dmtools.barplot(plotdata, [], "Test Plot", [], 'o', '', '', '', '')

def init_app (argv):
    "Initialise the application."
    global main_window
    main_window = libglade.GladeXML ("gdmplot.glade", "gdmplot_app")
    dic = {"on_about_activate"             : on_about_activate,
#          "on_browse_button_clicked"      : popup_file_selection,
           "on_open_button_clicked"        : popup_file_selection,
#	   "on_open_menu_item_activate"    : popup_file_selection,
	   "on_subset_checkbutton_toggled" : on_subset_checkbutton_toggled,
	   "on_axis_label_button_clicked"  : popup_axis_labels_dialog,
	   "on_clear_button_clicked"       : clear_all_plots,
	   "on_clear_menu_activate"        : clear_all_plots,
           "on_plot_button_clicked"        : generate_plot,
	   "on_generate_menu_activate"     : generate_plot,
           "on_app_destroy"                : quit_app,
           "on_quit_button_clicked"        : quit_app,
           "on_exit_activate"              : quit_app}
    main_window.signal_autoconnect (dic)

    # Check if any command line parametrs were given.  If so assume
    # the first is the filename and ignore the rest.

    if len (argv) > 1:
	    dataset_entry = main_window.get_widget ('dataset_entry')
	    dataset_entry.gtk_entry ().set_text (argv[1])
	    dataset_entry.gnome_entry ().append_history (1, argv[1])

    global xlabel
    xlabel = ""
    global ylabel
    ylabel = ""

def main ():
    init_app (sys.argv)
    gtk.mainloop ()

if __name__ == '__main__': main()
